// Author: A.Voss@FH-Aachen.de
//
// move ctor, move assignment
// tags: operator=

#include <iostream>
#include <cstdlib>
#include <string>
#include <cmath>
using namespace std;

#include "../../include/tools.h"

#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"


// we have rvalues and lvalues (locator value):
// - a lvalue is something whose address can be taken, one can assign something 
//   to lvalues;
// - rvalues are non-lvalues, or, every expression that results in a temporary object

// for more types (xvalue etc.) see
// http://en.cppreference.com/w/cpp/language/value_category


// to observe temp. objects we disable an optimization technique 
// called copy elision 

//#pragma GCC optimize "-fno-elide-constructors"

//-fno-elide-constructors
// The C++ standard allows an implementation to omit creating a 
// temporary which is only used to initialize another object of 
// the same type. Specifying this option disables that optimization, 
// and forces G++ to call the copy constructor in all cases. 


// a global variable (it has an address)
int m = 42;

// a ref to the global var m
int & set_m() { return m; }

// a temp. result-object
string get_name() { return "N.N."; }

// lvalue variant (also accepts rvalues, comment out function f below)
void f(const string& s)  { 
    tools_log(); cout << "ref  f& str " << s << endl; 
}

// rvalue variant (rvalues only)
void f(const string&& s) { 
    tools_log(); cout << "move f&& str " << s << endl; 
}

class C {
public:
    int n = 0;
    
    C(int N) : n(N) { 
        tools_log(); cout << "ctor C (" << n << ")" << endl; 
    }
    
    C(const C & c) : n(c.n) { 
        tools_log(); cout << "ctor C(&) (" << n << ")" << endl; 
    }
    
    C & operator=(const C & c) {
        n = c.n; 
        tools_log(); cout << "op=(&) (" << n << ")" << endl; return *this; 
    }
    
    C & operator+=(const C & c) { 
        n += c.n;
        return *this; 
    }

    C & operator*=(const C & c) { 
        n *= c.n;
        return *this; 
    }
    
    friend C operator+(const C & lhs, const C & rhs) { 
        tools_log(); cout << "op+ (" << lhs.n << "," << rhs.n << ")" << endl; 
        C result(lhs);
        result += rhs;
        return result; 
    }

    friend C operator*(C lhs, const C & rhs) { 
        tools_log(); cout << "op* (" << lhs.n << "," << rhs.n << ")" << endl; 
        lhs*=rhs; 
        return lhs; 
    }

    // comment out the following two functions to see how many temp. objs are created
    
    // this is the move-constructor
    C(const C && c) { 
        n = c.n;
        tools_log(); cout << "ctor C(&&) (" << n << ")" << endl; 
    }

    // this is the move-assignment-operator
    C & operator=(const C && c) {
        n = c.n;
        tools_log(); cout << "op=(&&) (" << n << ")" << endl; 
        return *this; 
    }
    
};


int main()
{
    int n;

    tools_log(); cout << "start    n=" << n << ", m=" << m << endl;

    n = 1;          // n is an lvalue
    set_m() = 23;   // lvalue that is not a variable
    tools_log(); cout << "assigned n=" << n << ", m=" << m << endl;
    
    string s;
    s = get_name(); // get_name is an rvalue
    
    // call different f's
    f(s);
    f(get_name());


    // compare number of temp. objects with or without move-ops.
    C c1(1),c2(2),c3(3);

    tools_log(); cout << "c1=c2+c3" << endl;
    c1.operator=(operator+(c2,c3)); // long version of c1=c2+c3
    cout << endl;

    c1 = c2+c3;
    cout << endl;
    
    c1 = c2*c3;
    cout << endl;

    // tools_log(); cout << "c1=c2+c3+c2" << endl;
    // c1 = c2+c3+c2;

    return EXIT_SUCCESS;
}

